searchData={"items":[{"type":"module","title":"Flint","doc":"Declarative [`Ecto`](https://github.com/elixir-ecto/ecto) `embedded_schema`s for data validation, coercion, and manipulation.","ref":"Flint.html"},{"type":"module","title":"Feature Overview - Flint","doc":"* `!` Variants of Ecto `field`, `embeds_one`, and `embeds_many` macros to mark a field as required ([Required Fields](#required-fields))\n* Colocated input transformations let you either transform input fields before validation or derive field values from other fields ([Derived Fields / Input Transformations](#derived-fields--input-transformations))\n* Colocated validations, so you can define common validations alongside field declarations ([Validations](#field-validations))\n* Colocated output transformations let you transform fields after validation ([Mappings / Output Transformations](#mappings--output-transformations))\n* Extensible using the `Flint.Extension` module. Default extensions include:\n  * `Accessible` - Adds `Access` implementation to the target schemas\n  * `JSON` - Adds a custom JSON encoding (`Jason` and `Poison` supported) implementation to the target schemas\n  * `Embedded` - Sets good default module attribute values used by `Ecto` specifically tailored for in-memory embedded schemas\n  * And more!\n* New [`Ecto.Schema` Reflection Functions](https://hexdocs.pm/ecto/Ecto.Schema.html#module-reflection)\n  * `__schema__(:required)` - Returns list of fields marked as required (from `!` macros)\n  * And more!\n* Convenient generated function (`changeset`,`new`,`new!`,...) ([Generated Functions](#generated-functions))\n* Configurable `Application`-wide defaults for `Ecto.Schema` API ([Config](#config))\n* Conveniently create new `Ecto` types using the `Flint.Type` module and its  `deftype/2` macro ([`Flint.Type`](#flinttype))","ref":"Flint.html#module-feature-overview"},{"type":"module","title":"Installation - Flint","doc":"```elixir\ndef deps do\n  [\n    {:flint, \"~> 0.6\"},\n    # If you want access to the `Typed` extension to add generated typespecs\n    {:typed_ecto_schema, \"~> 0.4\", runtime: false}\n  ]\nend\n```","ref":"Flint.html#module-installation"},{"type":"module","title":"Motivation - Flint","doc":"`Flint` is built on top of Ecto and is meant to provide good defaults for using `embedded_schema`s for use outside of a database.\nIt also adds a bevy of convenient features to the existing `Ecto` API to make writing schemas and validations much quicker.\n\nOf course, since you're using Ecto, you can use this for use as an ORM, but this is emphasizing the use of `embedded_schema`s as just more expressive and powerful maps while keeping compatibility with `Ecto.Changeset`, `Ecto.Type`, and all of the other benefits Ecto has to offer.\n\nIn particular, `Flint` focuses on making it more ergonomic to use `embedded_schema`s as a superset of Maps, so a `Flint.Schema` by default implements the `Access` behaviour and implements the `Jason.Encoder` protocol.\n\n`Flint` also was made to leverage the distinction `Ecto` makes between the embedded representation of the schema and the dumped representation. This means that you can dictate how you want the Elixir-side representation to look, and then provide transformations\nfor how it should be dumped, which helps when you want the serialized representation to look different.\n\nThis is useful if you want to make changes in the server-side code without needing to change the client-side (or vice-versa). Or perhaps you want a mapped representation, where instead of an `Ecto.Enum` just converting its atom key to a string when dumped, it gets mapped to an integer, etc.","ref":"Flint.html#module-motivation"},{"type":"module","title":"Basic Usage - Flint","doc":"If you want to declare a schema with `Flint`, just `use Flint.Schema` within your module, and now you have access to `Flint`'s implementation of the\n`embedded_schema/1` macro.  You can declare an `embedded_schema` within your module as you otherwise would with `Ecto`. Within the `embedded_schema/1` block, you also have access to `Flint`s implementations of `embeds_one`,`embeds_one!`,`embeds_many`, `embeds_many!`, `field`, and `field!`.\n\n```elixir\ndefmodule User do\n  use Flint.Schema\n\n  embedded_schema do\n    field! :username, :string\n    field! :password, :string, redacted: true\n    field :nickname, :string\n  end\nend\n```\n\n## `Flint` Types\n\n`Flint` also comes with some types that are automatically aliased when you `use Flint`.\n\n### `Union`\n\nUnion type for Ecto. Allows the field to be any of the specified types.\n\n## `Flint.Type`\n\n`Flint.Type` is meant to make writing new `Ecto` types require much less boilerplate, because you can base your type off of an existing type and only modify the callbacks that have different behavior.\n\nSimply `use Flint.Type` and pass the `:extends` option which says which type module to inherit callbacks\nfrom.  This will delegate all required callbacks and any implemented optional callbacks and make them\noverridable.\n\nIt also lets you make a type from an `Ecto.ParameterizedType` with default parameter values.\nYou may supply any number of default parameters. This essentially provides a new\n`init/1` implementation for the type, supplying the default values, while not affecting any of the\nother `Ecto.ParameterizedType` callbacks. You may still override the newly set defaults at the local level.\n\nJust supply all options that you wish to be defaults as extra options when using `Flint.Type`.\n\nYou may override any of the inherited callbacks inherity from the extended module\nin the case that you wish to customize the module further.\n\n#","ref":"Flint.html#module-basic-usage"},{"type":"module","title":"Examples - Flint","doc":"``` elixir\ndefmodule Category do\n  use Flint.Type, extends: Ecto.Enum, values: [:folder, :file]\nend\n```\n\nThis will apply default `values` to `Ecto.Enum` when you supply a `Category` type\nto an Ecto schema. You may still override the values if you supply the `:values`\noption for the field.\n\n```elixir\nimport Flint.Type\ndeftype NewUID, extends: Ecto.UUID, dump: &String.length/1\n```\n\nThis will create a new `NewUID` type that behaves exactly like an `Ecto.UUID` except it dumps\nits string length.","ref":"Flint.html#module-examples"},{"type":"module","title":"Generated Functions - Flint","doc":"`Flint` provides default implementations for the following functions for any schema declaration. Each of these is overridable.\n\n* `changeset` - Creates a changeset by casting all fields and validating all that were marked as required. If a `:default` key is provided for a field, then any use of a bang (`!`) declaration will essentially be ignored since the cast will fall back to the default before any validations are performed.\n* `new` - Creates a new changeset from the empty module struct and applies the changes (regardless of whether the changeset was valid).\n* `new!` - Same as new, except raises if the changeset is not valid.","ref":"Flint.html#module-generated-functions"},{"type":"module","title":"Flint Core - Flint","doc":"The core of Flint is the additional schema macros, which includes the bang (`!`) variants to mark\na field as required, and the added support of validations through `do` blocks to fields, as well\nas the `Flint.Extension` API that allows extensions to define additional acceptable `field` options\nand module attributes that can be reflected upon.\n\nAll other functionality comes in the form of `Flint` extensions.\n\nAt their core, the new `field` and `field!` macros' only additional functionality over the default `Ecto` macros is to store the allowed `Flint` options into module attributes which are exposed as new reflection functions.\n\nThe bulk of the work done in Flint with validations and transformations of data occurs in the generated `changeset` function, which leaves it up to the end user whether to use the default implementation, roll their own from scratch using the information exposed through the reflection functions, or do something in between (such as tuning which extensions you use).\n\nWhen you `use Flint.Schema`, you declare an overridable `changeset` function for your schema module that by default just\ndelegates to the `Flint.Changeset.changeset/3` function.\n\nThe `Flint.Changeset.changeset/3` function operates as the following pipeline:\n\n1. Cast all fields (including embeds)\n2. Validate required fields ([Required Fields](#required-fields))\n\n#","ref":"Flint.html#module-flint-core"},{"type":"module","title":"Required Fields - Flint","doc":"`Flint` adds the convenience bang (`!`) macros (`embed_one!`,`embed_many!`, `field!`) for field declarations within your struct to declare a field as required within its `changeset` function.\n\n`Flint` schemas also have a new reflection function in addition to the normal [`Ecto` reflection functions](https://hexdocs.pm/ecto/Ecto.Schema.html#module-reflection).\n\n* `__schema__(:required)` -- Returns a list of all fields that were marked as required.\n\n#","ref":"Flint.html#module-required-fields"},{"type":"module","title":"Field Validations - Flint","doc":"#### `field` `do` Blocks\n\nIn `Flint`, the `field` and `field!` macros now accept an optional `do` block to define condition/error pairs.\n\n```elixir\n  embedded_schema do\n    field! :type, :string do\n      type not in ~w[elf human] -> \"Expected elf or human, got: #{type}\"\n    end\n\n    field! :age, :integer do\n      age < 0 ->\n        \"Nobody can have a negative age\"\n\n      type == \"elf\" and age > max_elf_age ->\n        \"Attention! The elf has become a bug! Should be dead already!\"\n\n      type == \"human\" and age > max_human_age ->\n        \"Expected human to have up to #{max_human_age}, got: #{age}\"\n    end\n  end\n```\n\n```elixir\nmax_elf_age = 400\nmax_human_age = 120\nCharacter.new!(%{type: \"elf\", age: 10}, binding())\n```\n\n```elixir\n** (ArgumentError) %Character{type: [\"Expected elf or human, got: orc\"], age: 10}\n    (flint 0.0.1) lib/schema.ex:617: Flint.Schema.new!/3\n    (elixir 1.15.7) src/elixir.erl:396: :elixir.eval_external_handler/3\n    (stdlib 5.1.1) erl_eval.erl:750: :erl_eval.do_apply/7\n    (elixir 1.15.7) src/elixir.erl:375: :elixir.eval_forms/4\n    (elixir 1.15.7) lib/module/parallel_checker.ex:112: Module.ParallelChecker.verify/1\n    lib/livebook/runtime/evaluator.ex:622: anonymous fn/3 in Livebook.Runtime.Evaluator.eval/4\n    (elixir 1.15.7) lib/code.ex:574: Code.with_diagnostics/2\n```\n\nThe `:do` block accepts a list of validation clauses, where is clause is of the form:\n\n`failure condition -> Error Message`\n\nIn the `:do` block expressions, the same rules apply as mentioned across this documentation. You can pass\nbindings to apply to the expression, and field name bindings will automatically be passed to the expression so you can just use the field names as variables.\n\nAdditionally, you will have access to all local functions and imported functions that the parent module would have, so you can write expressions as you would in the parent module.\n\nThe AST representations of these vaildations are stored in a module attribute and can be retrieved using\n\n```elixir\n__schema__(:blocks)\n```\n\nThese validations are checked in the default `Flint.Changeset.changeset`.\n\n##","ref":"Flint.html#module-field-validations"},{"type":"module","title":"Parameterized Validations - Flint","doc":"You can even parameterize the options passed to the validations:\n\n```elixir\ndefmodule Person do\n  use Flint.Schema\n\n  embedded_schema do\n    field! :first_name, :string,  max: 10, min: 5\n    field! :last_name, :string, min: 5, max: 10\n    field :favorite_colors, {:array, :string}, subset_of: [\"red\", \"blue\", \"green\"]\n    field! :age, :integer, greater_than: 0, less_than: max_age\n  end\nend\n```\n\nIf you do this, make sure to pass the options as a keyword list into the call to `changeset`:\n\n```elixir\nPerson.changeset(\n  %Person{},\n  %{first_name: \"Bob\", last_name: \"Smith\", favorite_colors: [\"red\", \"blue\", \"pink\"], age: 101},\n  [max_age: 100]\n)\n```\n\n```elixir\n#Ecto.Changeset ,\n  valid?: false,\n  ...\n>\n```\n\n##","ref":"Flint.html#module-parameterized-validations"},{"type":"module","title":"Validate With Respect to Other Fields - Flint","doc":"You might find yourself wishing to validate a field conditionally based on the values of other fields. In `Flint`, you\ncan do this with any validation! Since all validations already accept parameterized conditions, they also let you refer\nto previously defined fields declared with `field` or `field!` macros. Just use a variable of the same name as the field(s) you want to refer to, and they will be bound to their respective variables.","ref":"Flint.html#module-validate-with-respect-to-other-fields"},{"type":"module","title":"Extensions - Flint","doc":"Flint provides an extensible architecture using the `Flint.Extension` module.\n\nIn fact, most of the core features that `Flint` offers are written as `Flint` extensions.\n\nThe default extensions can be retrieved using `Flint.default_extensions()`\n\n```elixir\nFlint.default_extensions()\n\n[\n  Flint.Extensions.PreTransforms,\n  Flint.Extensions.When,\n  Flint.Extensions.EctoValidations,\n  Flint.Extensions.PostTransforms,\n  Flint.Extensions.Accessible,\n  Flint.Extensions.Embedded,\n  Flint.Extensions.JSON\n]\n```\n\nTo use extensions, you can specify them when using `Flint.Schema` in your module. If the `:extensions` option\nis not provided, the default extensions will be used.\n\n```elixir\ndefmodule MySchema do\n  use Flint.Schema,\n    extensions: [Accessible, JSON]\n\n  embedded_schema do\n    # Schema fields...\n  end\nend\n```\n\n**Note that you don't have to write the fully-qualified name for modules in the `Flint.Extensions` module.**\n\nYou can use `Flint.default_extensions()` to refer to the default extensions, which you will have to\nexplicitly add if you pass custom values to the `:extensions` option when using `Flint.Schema`, eg.\n\n```elixir\ndefmodule MySchema do\n  use Flint.Schema,\n    extensions: Flint.default_extensions() ++ [MyExtension]\n\n  embedded_schema do\n    # Schema fields...\n  end\nend\n```\n\nYou can also create custom extensions by `use`ing `Flint.Extension`. This allows you to add additional functionality or modify the behavior of Flint schemas according to your specific needs.\n\nFor more details on creating and using extensions, refer to the `Flint.Extension` module documentation.","ref":"Flint.html#module-extensions"},{"type":"module","title":"Aliases - Flint","doc":"If you don't like the name of an option, you can provide a compile-time list of aliases to map new option names to existing options\n([Validation Options](#basic-validation-options) and [Transformation Options](#input-and-output-transformations)).\n\nIn your config, add an `:aliases` key with a `Keyword` value, where each key is the new alias, and the value is an existing option name.\n\nFor example, these are default aliases implemented in `Flint`:\n\n```elixir\nconfig Flint, aliases: [\n    lt: :less_than,\n    gt: :greater_than,\n    le: :less_than_or_equal_to,\n    ge: :greater_than_or_equal_to,\n    eq: :equal_to,\n    ne: :not_equal_to\n  ]\n```\n\n**NOTE** If you add your own aliases and want to keep these above defaults, you will have to add them manually to your aliases.","ref":"Flint.html#module-aliases"},{"type":"module","title":"Config - Flint","doc":"You can configure the default options set by `Flint`.\n\n* `embeds_one`: The default arguments when using `embeds_one`. Defaults to `[defaults_to_struct: true, on_replace: :delete]`\n* `embeds_one!`: The default arguments when using `embeds_one!`. Defaults to `[on_replace: :delete]`\n* `embeds_many`: The default arguments when using `embeds_many` or `embeds_many!`. Defaults to `[on_replace: :delete]`\n* `embeds_many!`: The default arguments when using `embeds_many!`. Defaults to `[on_replace: :delete]`\n* `:aliases`: [Aliases](#aliases)\n\nYou can also configure any aliases you want to use for schema validations.","ref":"Flint.html#module-config"},{"type":"module","title":"Embedded vs Dumped Representations - Flint","doc":"`Flint` takes advantage of the distinction `Ecto` makes between an `embedded_schema`'s embedded and dumped representations.\n\nFor example, `Flint` provides the `Flint.Types.Enum` type, which are `Ecto.Enum`s where, when given values that are `Keyword` (rather than just lists of atoms) will have their keys\nbe the embedded representation and will have the values be the dumped representation.\n\n```elixir\ndefmodule Book do\n  use Flint.Schema, schema: [\n    field(:genre, Flint.Types.Enum, values: [biography: 0, science_fiction: 1, fantasy: 2, mystery: 3])\n  ]\nend\n\nbook = Book.new(%{genre: \"biography\"})\n# %Book{genre: :biography}\n\nFlint.Schema.dump(book)\n# %{genre: 0}\n```\n\nIn this example, you can see how you can share multiple representations of the same data using this distinction.\n\nYou can also implement your own `Ecto.Type` and further customize this (see [`Flint.Type`](#flinttype)):\n\n```elixir\ndefmodule ContentType do\n  use Ecto.Type\n  def type, do: :atom\n\n  def cast(\"application/json\"), do: {:ok, :json}\n\n  def cast(_), do: :error\n  def load(_), do: :error\n\n  def dump(:json), do: {:ok, \"application/json\"}\n  def dump(_), do: :error\n\n  def embed_as(_) do\n    :dump\n  end\nend\n```\n\nHere, `cast` will be called when creating a new `Flint` schema from a map, and `dump` will be used\nwhen calling `Flint.Schema.dump/1`.\n\n```elixir\ndefmodule URL do\n  use Flint, schema: [\n    field(:content_type, ContentType)\n  ]\nend\n```\n\n```elixir\nurl = URL.new(%{content_type: \"application/json\"})\n# %URL{content_type: :json}\n\nFlint.Schema.dump(url)\n# %{content_type: \"application/json\"}\n```","ref":"Flint.html#module-embedded-vs-dumped-representations"},{"type":"module","title":"Examples - Flint","doc":"You can view the [Notebooks folder](https://github.com/acalejos/flint/tree/main/notebooks) for some examples in Livebook.\n\nYou can also look at [Merquery](https://github.com/acalejos/merquery/tree/main/lib/merquery/schemas) for a real, comprehensive\nexample of how to use `Flint`.","ref":"Flint.html#module-examples"},{"type":"function","title":"Flint.default_extensions/1","doc":"","ref":"Flint.html#default_extensions/1"},{"type":"module","title":"Flint.Changeset","doc":"The base `changeset` function defined by `Flint`. `Flint.Changeset` uses the module attributes\nthat are collected when using the `Flint.Schema` macros to perform transformations and validations.","ref":"Flint.Changeset.html"},{"type":"function","title":"Flint.Changeset.changeset/3","doc":"Given a `Flint` (or `Ecto`) schema and params (can be a map, struct of the given schema, or an existing changeset),\napplies all steps of the `Flint.Changeset` to generate a new changeset.\n\nThis function casts all fields (recursively casting all embeds using this same function),\nvalidates required fields (specified using the bang (`!`) macros exposed by `Flint`),\noutputting the resulting `Ecto.Changeset`.","ref":"Flint.Changeset.html#changeset/3"},{"type":"module","title":"Flint.Schema","doc":"`Flint.Schema` provides custom implementations of certain `Ecto` `embedded_schema` DSL keywords.\n\nWhen you `use Flint`, all of these definitions are imported into the module and override the default\n`Ecto.Schema` implementations. You should not have to directly interact with this module.","ref":"Flint.Schema.html"},{"type":"function","title":"Flint.Schema.dump/1","doc":"Dumps the JSON representation of the given schema","ref":"Flint.Schema.html#dump/1"},{"type":"macro","title":"Flint.Schema.embedded_schema/1","doc":"Wraps `Ecto`'s `embedded_schema` macro, injecting `Flint`'s custom macro implementation into the module space.","ref":"Flint.Schema.html#embedded_schema/1"},{"type":"macro","title":"Flint.Schema.embeds_many/3","doc":"Wraps `Ecto`'s `embeds_many` macro to accept additional options which are consumed by `Flint.Changeset`.\n\n`Flint` options that are passed to `embeds_many` are stripped by `Flint` before passing them to `Ecto.Schema.embeds_many` and\nstored in module attributed for the schema's module.\n\nThe following default options are passed to `Flint.Schema.embeds_many` and can be overriden at the `Application` level or at the\nlocal call level.","ref":"Flint.Schema.html#embeds_many/3"},{"type":"macro","title":"Default Options - Flint.Schema.embeds_many/3","doc":"```elixir\non_replace: :delete\n```","ref":"Flint.Schema.html#embeds_many/3-default-options"},{"type":"macro","title":"Flint.Schema.embeds_many/4","doc":"","ref":"Flint.Schema.html#embeds_many/4"},{"type":"macro","title":"Flint.Schema.embeds_many!/3","doc":"Marks an embeds_many field as required by storing metadata in a module attribute, then calls `Flint.Schema.embeds_many`.\n\nThe metadata tracking is not enforced at the schema / struct level (eg. through enforced keys), but rather\nat the schema validation level (through something such as `Flint.Changeset.changeset`).\n\nThe following default options are passed to `Flint.Schema.embeds_many!` and can be overriden at the `Application` level or at the\nlocal call level.","ref":"Flint.Schema.html#embeds_many!/3"},{"type":"macro","title":"Default Options - Flint.Schema.embeds_many!/3","doc":"```elixir\non_replace: :delete\n```","ref":"Flint.Schema.html#embeds_many!/3-default-options"},{"type":"macro","title":"Flint.Schema.embeds_many!/4","doc":"","ref":"Flint.Schema.html#embeds_many!/4"},{"type":"macro","title":"Flint.Schema.embeds_one/3","doc":"Wraps `Ecto`'s `embeds_one` macro to accept additional options which are consumed by `Flint.Changeset`.\n\n`Flint` options that are passed to `embeds_one` are stripped by `Flint` before passing them to `Ecto.Schema.embeds_one` and\nstored in module attributed for the schema's module.\n\nThe following default options are passed to `Flint.Schema.embeds_one` and can be overriden at the `Application` level or at the\nlocal call level.","ref":"Flint.Schema.html#embeds_one/3"},{"type":"macro","title":"Default Options - Flint.Schema.embeds_one/3","doc":"```elixir\ndefaults_to_struct: true,\non_replace: :delete\n```","ref":"Flint.Schema.html#embeds_one/3-default-options"},{"type":"macro","title":"Flint.Schema.embeds_one/4","doc":"","ref":"Flint.Schema.html#embeds_one/4"},{"type":"macro","title":"Flint.Schema.embeds_one!/3","doc":"Marks an embeds_one field as required by storing metadata in a module attribute, then calls `Flint.Schema.embeds_one`.\n\nThe metadata tracking is not enforced at the schema / struct level (eg. through enforced keys), but rather\nat the schema validation level (through something such as `Flint.Changeset.changeset`).\n\nThe following default options are passed to `Flint.Schema.embeds_one!` and can be overriden at the `Application` level or at the\nlocal call level.","ref":"Flint.Schema.html#embeds_one!/3"},{"type":"macro","title":"Default Options - Flint.Schema.embeds_one!/3","doc":"```elixir\non_replace: :delete\n```","ref":"Flint.Schema.html#embeds_one!/3-default-options"},{"type":"macro","title":"Flint.Schema.embeds_one!/4","doc":"","ref":"Flint.Schema.html#embeds_one!/4"},{"type":"macro","title":"Flint.Schema.field/3","doc":"Wraps `Ecto`'s `field` macro to accept additional options which are consumed by `Flint.Changeset`.\n\n`Flint` options that are passed to `field` are stripped by `Flint` before passing them to `Ecto.Schema.field` and\nstored in module attributed for the schema's module.\n\nIf no `Flint`-specific options or features are used, this acts the same as `Ecto.Schema.field`.","ref":"Flint.Schema.html#field/3"},{"type":"macro","title":"Flint.Schema.field!/3","doc":"Marks a field as required by storing metadata in a module attribute, then calls `Flint.Schema.field`.\n\nThe metadata tracking is not enforced at the schema / struct level (eg. through enforced keys), but rather\nat the schema validation level (through something such as `Flint.Changeset.changeset`).","ref":"Flint.Schema.html#field!/3"},{"type":"function","title":"Flint.Schema.new/3","doc":"Creates a new schema struct according to the schema's `changeset` implementation, immediately applying\nthe changes from the changeset regardless of errors.\n\nIf you want to manually handle error cases, you should use the `changeset` function itself.","ref":"Flint.Schema.html#new/3"},{"type":"function","title":"Flint.Schema.new!/3","doc":"Same as `new`, except will `raise` if any errors exist in the changeset.","ref":"Flint.Schema.html#new!/3"},{"type":"module","title":"Flint.Type","doc":"`Flint.Type` is meant to make writing new `Ecto` types require much less boilerplate, because you can base your\ntype off of an existing type and only modify the callbacks that have different behavior.\n\nSimply `use Flint.Type` and pass the `:extends` option which says which type module to inherit callbacks\nfrom.  This will delegate all required callbacks and any implemented optional callbacks and make them\noverridable.\n\nIt also lets you make a type from an `Ecto.ParameterizedType` with default parameter values.\nYou may supply any number of default parameters. This essentially provides a new\n`init/1` implementation for the type, supplying the default values, while not affecting any of the\nother `Ecto.ParameterizedType` callbacks. You may still override the newly set defaults at the local level.\n\nJust supply all options that you wish to be defaults as extra options when using `Flint.Type`.\n\nYou may override any of the inherited callbacks inherity from the extended module\nin the case that you wish to customize the module further.","ref":"Flint.Type.html"},{"type":"module","title":"Examples - Flint.Type","doc":"```elixir\ndefmodule Category do\n  use Flint.Type, extends: Ecto.Enum, values: [:folder, :file]\nend\n```\n\nThis will apply default `values` to `Ecto.Enum` when you supply a `Category` type\nto an Ecto schema. You may still override the values if you supply the `:values`\noption for the field.\n\n```elixir\nimport Flint.Type\ndeftype NewUID, extends: Ecto.UUID, dump: &String.length/1\n```\n\nThis will create a new `NewUID` type that behaves exactly like an `Ecto.UUID` except it dumps\nits string length.","ref":"Flint.Type.html#module-examples"},{"type":"macro","title":"Flint.Type.deftype/2","doc":"A shorthand for creating a new `Flint.Type`. This is equivalent to\ncreating a new module with name `module` and calling `use Flint.Type` and passing `opts`.","ref":"Flint.Type.html#deftype/2"},{"type":"macro","title":"Example - Flint.Type.deftype/2","doc":"```elixir\ndeftype Vehicle, extends: Ecto.Enum, values: [:car, :motorcycle, :truck]\n```","ref":"Flint.Type.html#deftype/2-example"},{"type":"module","title":"Flint.Types.Enum","doc":"","ref":"Flint.Types.Enum.html"},{"type":"function","title":"Flint.Types.Enum.base_type/0","doc":"","ref":"Flint.Types.Enum.html#base_type/0"},{"type":"function","title":"Flint.Types.Enum.cast/2","doc":"","ref":"Flint.Types.Enum.html#cast/2"},{"type":"function","title":"Flint.Types.Enum.dump/3","doc":"","ref":"Flint.Types.Enum.html#dump/3"},{"type":"function","title":"Flint.Types.Enum.format/1","doc":"","ref":"Flint.Types.Enum.html#format/1"},{"type":"function","title":"Flint.Types.Enum.init/1","doc":"","ref":"Flint.Types.Enum.html#init/1"},{"type":"function","title":"Flint.Types.Enum.load/3","doc":"","ref":"Flint.Types.Enum.html#load/3"},{"type":"function","title":"Flint.Types.Enum.type/1","doc":"","ref":"Flint.Types.Enum.html#type/1"},{"type":"module","title":"Flint.Types.Union","doc":"Union type for Ecto. Allows the field to be any of the specified types.\n\nThis is particularly useful if you want to allow one of multiple types and preserve the type of the\ninput, so long as it adheres to the type (rather than forcing a cast).\n\nFor example, suppose you want to allow an input to be a float, or an integer, but you want to\npreserve whichever type is passed (keep floats as floats and keep integers as integers), you could\nuse a `Union`.","ref":"Flint.Types.Union.html"},{"type":"module","title":"Options - Flint.Types.Union","doc":"* `:oneof` - Allowed types. Can be any valid `Ecto` type (including custom types). Required.\n* `:eager` - Whether to eagerly cast the field value into the `:oneof` in the order that\nthey appear in the list. If `false`, `Flint` will attempt to determine if the value maps to a\nvalid `Ecto` type before casting and will prefer that type (if it appears in `:oneof`). Defaults\nto `false`.","ref":"Flint.Types.Union.html#module-options"},{"type":"module","title":"Example - Flint.Types.Union","doc":"```elixir\ndefmodule Character do\n  use Flint\n\n  embedded_schema do\n    field! :name, :string\n    field! :strength, Union, oneof: [:integer, :float]\n  end\nend\n```\n\n```elixir\nCharacter.new!(%{name: \"Elf\", strength: 10})\n```\n\n```\n%Character{name: \"Elf\", strength: 10}\n```\n\n```elixir\nCharacter.new!(%{name: \"Elf\", strength: 10.0})\n```\n\n```\n%Character{name: \"Elf\", strength: 10.0}\n```","ref":"Flint.Types.Union.html#module-example"},{"type":"behaviour","title":"Flint.Extension","doc":"`Flint` extensions allow developers to easily hook into `Flint` metaprogramming lifecycle to add extra data into the embedded\nschema reflection functions.\n\nFlint currently offers four ways to extend behavior:\n\n1. Schema-level attributes\n2. Field-level additional options\n3. Default `embedded_schema` definitions\n4. Injected Code\n\nExtension authors define what fields / options / attributes `Flint` should look for in the module / schema definition and\nstrip out and store in a schema reflection function, but it is still the resposibility of either the extension author or\nthe end user to make use of the stored information.","ref":"Flint.Extension.html"},{"type":"behaviour","title":"Schema-Level Attributes - Flint.Extension","doc":"These are simply module attributes that are pre-registered with `Flint`, and can be given a default value\nas well as a validation function. When you use an extension that registers an attribute, then a new `__schema__`\nreflection function is added for each attribute name, with the attribute name as the argument.\n\n**Note that the validation occurs at compile time**\n\n#","ref":"Flint.Extension.html#module-schema-level-attributes"},{"type":"behaviour","title":"Example - Flint.Extension","doc":"Given the following extension:\n\n```elixir\ndefmodule Returnable do\n  use Flint.Extension\n\n  attribute :returns, validator: fn returns -> is_binary(returns) end\nend\n```\n\nAnd the schema\n\n```elixir\ndefmodule Schema do\n  use Flint.Schema, extensions: [Returnable]\n  @returns \"something\"\n  embedded_schema do\n    ...\n  end\nend\n```\n\nThen you can reflect on this new attribute with:\n\n```elixir\nSchema.__schema__(:returns)\n```","ref":"Flint.Extension.html#module-example"},{"type":"behaviour","title":"Field-Level Options - Flint.Extension","doc":"You can also register additional field-level keyword options to be consumed in a downstream function.\n\nThese function similarly to the built-in extra options that `Flint` provides, where the options are\nstripped and stored in a module attribute (and subsequently in a `__schema__` reflection function)\nbefore passing the valid `Ecto.Schema` options to `Ecto` itself.\n\n**Note that the validation occurs at compile time**\n\n#","ref":"Flint.Extension.html#module-field-level-options"},{"type":"behaviour","title":"Example - Flint.Extension","doc":"Given the following extension that enables Go-like JSON marshalling options:\n\n```elixir\ndefmodule JSON do\n  use Flint.Extension\n\n  option :name, required: false, validator: &is_binary/1\n  option :omitempty, required: false, default: false, validator: &is_boolean/1\n  option :ignore, required: false, default: false, validator: &is_boolean/1\nend\n```\n\nAnd the following schema:\n\n```elixir\ndefmodule Schema do\n  use Flint.Schema, extensions: [JSON]\n  embedded_schema do\n    field :myfield, :string, name: \"my_field\", omitempty: true\n  end\nend\n```\n\nThen you can access these specific fields with:\n\n```elixir\nSchema.__schema__(:extra_options)\n```\n\n```elixir\n[\n  myfield: [ignore: false, omitempty: true, name: \"my_field\"],\n]\n```","ref":"Flint.Extension.html#module-example"},{"type":"behaviour","title":"Default `embedded_schema` definitions - Flint.Extension","doc":"Sometimes you might want to always have one of more fields defined for a given schema type without having\nto write it each time. Much in the same way that, by default, the `@prmimary_key` attribute will set an `:id` field\nfor the schema, you might wish to template out some fields and values.\n\nIf you want to define fields that are defined by default in a schema that uses your extension, you can use the\n`embedded_schema` macro within an extension and all `field`, `embeds_one` and `embeds_many` declarations will be\nmerged into the target schema that uses your extension.\n\nThese accept the same arguments as their counterparts in `Flint.Schema`, and will be\nadded at the end of the `embedded_schema` definition.\n\n> ##","ref":"Flint.Extension.html#module-default-embedded_schema-definitions"},{"type":"behaviour","title":"Duplicate Fields {: .info} - Flint.Extension","doc":">\n> Note that if you define a template for a field in your extension then anyone who uses your extension will be\n> unable to override that field name with their own definition.\n\n#","ref":"Flint.Extension.html#module-duplicate-fields-info"},{"type":"behaviour","title":"Example - Flint.Extension","doc":"Given the following extension:\n\n```elixir\ndefmodule Event do\n  use Flint.Extension\n\n  embedded_schema do\n    field! :timestamp, :utc_datetime_usec\n    field! :id, :binary_id\n  end\nend\n```\n\nand this schema:\n\n```elixir\ndefmodule Webhook do\n  use Flint.Schema, extensions: [Event, Embedded]\n\n  embedded_schema do\n    field :route, :string\n  end\nend\n```\n\nThen any schema that uses this extension will have these fields by default:\n\n```elixir\nWebhook.__schema__(:fields)\n# [:route, :timestamp, :id]\n```","ref":"Flint.Extension.html#module-example"},{"type":"behaviour","title":"Injected Code - Flint.Extension","doc":"Lastly, extensions allow you to define custom `__using__/1` macros which will be passed through\nto the target schema module. This is one of the core functionalities of extensions, and works the\nsame as you would normally `use` a module, and helps compartmentalize similar functionality.","ref":"Flint.Extension.html#module-injected-code"},{"type":"behaviour","title":"Default Extensions - Flint.Extension","doc":"By default, `Flint` will enable the following extensions:\n\n*  `Flint.Extensions.Block`\n*  `Flint.Extensions.Typed`\n*  `Flint.Extensions.PreTransforms`,\n*  `Flint.Extensions.When`,\n*  `Flint.Extensions.EctoValidations`,\n*  `Flint.Extensions.PostTransforms`,\n*  `Flint.Extensions.Accessible`,\n*  `Flint.Extensions.Embedded`,\n*  `Flint.Extensions.JSON`\n\nIf you want to pass your own list of extensions for a module, you will need to explicitly pass the defaults\nas well if you would like to keep them. You can use the convenience `Flint.default_extensions/0` constant\nif you want to include all of the defaults.\n\n\n#","ref":"Flint.Extension.html#module-default-extensions"},{"type":"behaviour","title":"Options - Flint.Extension","doc":"* `:extensions` (list of module that adopts `Spark.Dsl.Extension`) - A list of DSL extensions to add to the `Spark.Dsl`\n\n* `:otp_app` (`t:atom/0`) - The otp_app to use for any application configurable options\n\n* `:fragments` (list of `t:module/0`) - Fragments to include in the `Spark.Dsl`. See the fragments guide for more.","ref":"Flint.Extension.html#module-options"},{"type":"callback","title":"Flint.Extension.changeset/1","doc":"","ref":"Flint.Extension.html#c:changeset/1"},{"type":"callback","title":"Flint.Extension.changeset/2","doc":"","ref":"Flint.Extension.html#c:changeset/2"},{"type":"module","title":"Flint.Extensions.Accessible","doc":"An extension to automatically implement the `Access` behaviour for your struct,\ndeferring to the `Map` implementation.","ref":"Flint.Extensions.Accessible.html"},{"type":"module","title":"Flint.Extensions.Block","doc":"Adds support for `do` block in `field` and `field!` to add `validation_condition -> error_message` pairs to the field.\n\nBlock validations can be specified using `do` blocks in `field` and `field!`. These are specified as lists of `error_condition -> error_message` pairs. If the error condition returns `true`, then the corresponding `error_message` will be inserted into the changeset when using the generated `changeset`, `new`, and `new!` functions.\n\nWithin these validations, you can pass custom bindings, meaning that you can define these validations with respect to variables only available at runtime.\n\nIn addition to any bindings you pass, the calues of the fields themselves will be available as a variable with the same name as the field.\n\nYou can also refer to local and imported / aliased function within these validations as well.","ref":"Flint.Extensions.Block.html"},{"type":"function","title":"Flint.Extensions.Block.changeset/2","doc":"Uses the quoted expressions from the `Flint.Schema.field` and `Flint.Schema.field!`\n`do` blocks to validate the changeset.\n\nYou can optionally pass bindings to be added to the evaluation context.","ref":"Flint.Extensions.Block.html#changeset/2"},{"type":"macro","title":"Flint.Extensions.Block.embedded_schema/1","doc":"","ref":"Flint.Extensions.Block.html#embedded_schema/1"},{"type":"macro","title":"Flint.Extensions.Block.field/3","doc":"","ref":"Flint.Extensions.Block.html#field/3"},{"type":"macro","title":"Flint.Extensions.Block.field/4","doc":"","ref":"Flint.Extensions.Block.html#field/4"},{"type":"macro","title":"Flint.Extensions.Block.field!/3","doc":"","ref":"Flint.Extensions.Block.html#field!/3"},{"type":"macro","title":"Flint.Extensions.Block.field!/4","doc":"","ref":"Flint.Extensions.Block.html#field!/4"},{"type":"module","title":"Flint.Extensions.EctoValidations","doc":"Shorthand options for common validations found in `Ecto.Changeset`\n\nJust passthrough the option for the appropriate validation and this extension\nwill take care of calling the corresponding function from `Ecto.Changeset` on\nyour data.","ref":"Flint.Extensions.EctoValidations.html"},{"type":"module","title":"Options - Flint.Extensions.EctoValidations","doc":"* `:greater_than` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n* `:less_than` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n* `:less_than_or_equal_to` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n* `:greater_than_or_equal_to` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n* `:equal_to` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n* `:not_equal_to` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n* `:format` ([`Ecto.Changeset.validate_format/4`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_format/4))\n* `:subset_of` ([`Ecto.Changeset.validate_subset/4`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_subset/4))\n* `:in` ([`Ecto.Changeset.validate_inlusion/4`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_inclusion/4))\n* `:not_in` ([`Ecto.Changeset.validate_exclusion/4`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_exclusion/4))\n* `:is` ([`Ecto.Changeset.validate_length/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_length/3-options))\n* `:min` ([`Ecto.Changeset.validate_length/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_length/3-options))\n* `:max` ([`Ecto.Changeset.validate_length/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_length/3-options))\n* `:count` ([`Ecto.Changeset.validate_length/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_length/3-options))","ref":"Flint.Extensions.EctoValidations.html#module-options"},{"type":"module","title":"Aliases - Flint.Extensions.EctoValidations","doc":"By default, the following aliases are also available for convenience:\n\n```elixir\nconfig Flint, aliases: [\n  lt: :less_than,\n  gt: :greater_than,\n  le: :less_than_or_equal_to,\n  ge: :greater_than_or_equal_to,\n  eq: :equal_to,\n  ne: :not_equal_to\n]\n```","ref":"Flint.Extensions.EctoValidations.html#module-aliases"},{"type":"module","title":"Example - Flint.Extensions.EctoValidations","doc":"```elixir\ndefmodule Person do\n  use Flint.Schema\n\n  embedded_schema do\n    field! :first_name, :string,  max: 10, min: 5\n    field! :last_name, :string, min: 5, max: 10\n    field :favorite_colors, {:array, :string}, subset_of: [\"red\", \"blue\", \"green\"]\n    field! :age, :integer, greater_than: 0, less_than: max_age\n  end\nend\n```\n\n```elixir\nPerson.changeset(\n  %Person{},\n  %{first_name: \"Bob\", last_name: \"Smith\", favorite_colors: [\"red\", \"blue\", \"pink\"], age: 101},\n  [max_age: 100]\n)\n```\n\n```elixir\n#Ecto.Changeset ,\n  valid?: false,\n  ...\n>\n```","ref":"Flint.Extensions.EctoValidations.html#module-example"},{"type":"function","title":"Flint.Extensions.EctoValidations.changeset/2","doc":"Applies validations to each field according to the options passed in the schema specification.\n\nSee the `Field Validations` section of the README for more information on validation details.","ref":"Flint.Extensions.EctoValidations.html#changeset/2"},{"type":"module","title":"Flint.Extensions.Embedded","doc":"An extension to house common default configurations for embedded schemas. These configurations are specific for\nin-memory schemas.","ref":"Flint.Extensions.Embedded.html"},{"type":"module","title":"Attributes - Flint.Extensions.Embedded","doc":"The following attributes and defaults are set by this extension:\n\n* `:schema_prefix`\n* `:schema_context`\n* `:primary_key` - defaults to `false`\n* `:timestamp_opts` - defaults to `[type: :naive_datetime]`\n\nA new schema reflection function is made for each attribute:\n\n```elixir\n__schema__(:schema_context)\n...\n```","ref":"Flint.Extensions.Embedded.html#module-attributes"},{"type":"module","title":"Flint.Extensions.JSON","doc":"Provides JSON encoding capabilities for Flint schemas with Go-like marshalling options.\n\nThis extension enhances Flint schemas with customizable JSON serialization options,\nallowing fine-grained control over how fields are represented in JSON output.","ref":"Flint.Extensions.JSON.html"},{"type":"module","title":"Usage - Flint.Extensions.JSON","doc":"To use this extension, include it in your Flint schema:\n\n```elixir\ndefmodule MySchema do\n  use Flint.Schema,\n     extensions: [{JSON, json_module: :json}]  # Jason, or Poison\n     #extensions: [JSON] # (defaults to Jason if no args passed)\n\n  embedded_schema do\n    # Schema fields...\n  end\nend\n```","ref":"Flint.Extensions.JSON.html#module-usage"},{"type":"module","title":"JSON Encoding Options - Flint.Extensions.JSON","doc":"The following options can be specified for each field in your schema:\n\n- `:name` - Specifies a custom name for the field in the JSON output.\n- `:omitempty` - When set to `true`, omits the field from JSON output if its value is `nil`.\n- `:ignore` - When set to `true`, always excludes the field from JSON output.","ref":"Flint.Extensions.JSON.html#module-json-encoding-options"},{"type":"module","title":"Defining Options - Flint.Extensions.JSON","doc":"Options are defined directly in your schema using the `field` macro:\n\n```elixir\nembedded_schema do\n  field :id, :string, name: \"ID\"\n  field :title, :string, name: \"Title\", omitempty: true\n  field :internal_data, :map, ignore: true\nend\n```","ref":"Flint.Extensions.JSON.html#module-defining-options"},{"type":"module","title":"Example - Flint.Extensions.JSON","doc":"```elixir\ndefmodule Book do\n  use Flint.Schema,\n    extensions: [Embedded, Accessible, JSON]\n\n  embedded_schema do\n    field :id, :string, name: \"ISBN\"\n    field :title, :string\n    field :author, :string, omitempty: true\n    field :price, :decimal, name: \"SalePrice\"\n    field :internal_notes, :string, ignore: true\n  end\nend\n\nbook = %Book{\n  id: \"978-3-16-148410-0\",\n  title: \"Example Book\",\n  author: nil,\n  price: Decimal.new(\"29.99\"),\n  internal_notes: \"Not for customer eyes\"\n}\n\nJason.encode!(book)\n# Results in:\n# {\n#   \"ISBN\": \"978-3-16-148410-0\",\n#   \"title\": \"Example Book\",\n#   \"SalePrice\": \"29.99\"\n# }\n```\n\nNote how the `author` field is omitted due to `omitempty: true`, the `internal_notes`\nfield is ignored, and custom names are used for `id` and `price`.","ref":"Flint.Extensions.JSON.html#module-example"},{"type":"module","title":"Introspection - Flint.Extensions.JSON","doc":"You can inspect the JSON options for a schema using the `__schema__/1` function:\n\n```elixir\nBook.__schema__(:extra_options)\n# Returns:\n# [\n#   id: [name: \"ISBN\"],\n#   title: [],\n#   author: [omitempty: true],\n#   price: [name: \"SalePrice\"],\n#   internal_notes: [ignore: true]\n# ]\n```\n\nThis allows for runtime inspection and manipulation of JSON encoding behavior.","ref":"Flint.Extensions.JSON.html#module-introspection"},{"type":"module","title":"Implementation Details - Flint.Extensions.JSON","doc":"This extension implements the `Jason.Encoder` protocol for your schema, automatically\napplying the specified options during JSON encoding. It leverages Flint's schema\nintrospection capabilities to retrieve and apply the options.\n\nThe encoding process:\n1. Converts the struct to a map\n2. Applies the `:name` option to change key names\n3. Applies the `:omitempty` option to remove nil values\n4. Applies the `:ignore` option to exclude specified fields\n5. Encodes the resulting map to JSON\n\nThis approach provides a flexible and powerful way to control JSON serialization\ndirectly from your schema definition, promoting clean and maintainable code.","ref":"Flint.Extensions.JSON.html#module-implementation-details"},{"type":"module","title":"Flint.Extensions.PostTransforms","doc":"The `PostTransforms` extension adds the `:map` option to `Flint` schemas.\n\nThis works similarly to the `PreTransforms` extension, but uses the `:map` option rather than\nthe `:derive` option used by `PreTransforms`, and by default, applies to the field after all validations.\n\nThe same caveats apply to the `:map` expression as all other expressions, with the exception that the\n`:map` function **only** accepts arity-1 anonymous functions or non-anonymous function expressions\n(eg. using variable replacement).\n\nIn the following example, `:derived` is used to normalize incoming strings to downcase to prepare for\nthe validation, then the output is mapped to the uppercase string using the `:map` option.\n\n```elixir\ndefmodule Character do\n  use Flint.Schema\n\n  embedded_schema do\n    field! :type, :string, derive: &String.downcase/1, map: String.upcase(type) do\n      type not in ~w[elf human] -> \"Expected elf or human, got: #{type}\"\n    end\n\n    field! :age, :integer do\n      age < 0 ->\n        \"Nobody can have a negative age\"\n\n      type == \"elf\" and age > max_elf_age ->\n        \"Attention! The elf has become a bug! Should be dead already!\"\n\n      type == \"human\" and age > max_human_age ->\n        \"Expected human to have up to #{max_human_age}, got: #{age}\"\n    end\n  end\nend\n```\n\n```elixir\nmax_elf_age = 400\nmax_human_age = 120\nCharacter.new!(%{type: \"Elf\", age: 10}, binding())\n```\n\n```elixir\n%Character{type: \"ELF\", age: 10}\n```","ref":"Flint.Extensions.PostTransforms.html"},{"type":"function","title":"Flint.Extensions.PostTransforms.changeset/2","doc":"Applies transformations to each field according to the `:map` options passed in the schema specification.\n\nThese transformations are applied after validations when used within the default `Flint.Changeset.changeset` implementation.\n\nAccepts optional bindings which are passed to evaluated code.","ref":"Flint.Extensions.PostTransforms.html#changeset/2"},{"type":"module","title":"Flint.Extensions.PreTransforms","doc":"The `PreTransforms` provides a convenient `:derive` option to express how the field is computed.\n\n**By default, this occurs after casting and before validations.**\n\n`derived` fields let you define expressions with support for custom bindings to include any\n`field` declarations that occur before the current field.\n\n`:derive` will automatically put the result of the input expression into the field value.\nBy default, this occurs before any other validation, so you can still have access to `field`\nbindings and even the current computed field value (eg. within a `:when` validation from the\n`When` extension).\n\nYou can define a `derived` field with respect to the field itself, in which case it acts as\ntransformation. Typically in `Ecto`, incoming transformations of this support would happen\nat the `cast` step, which means the behavior is determined by the type in which you are casting into.\n`:derive` lets you apply a transformation after casting to change that behavior\nwithout changing the underlying allowed type.\n\nYou can also define a `derived` field with an expression that does not depend on the field,\nin which case it is suggested that you use the `field` macro instead of `field!` since any input\nin that case would be thrashed by the derived value. This means that a field can be completely\ndetermined as a product of other fields!\n\n```elixir\ndefmodule Test do\n  use Flint.Schema\n\n  embedded_schema do\n    field! :category, Union, oneof: [Ecto.Enum, :decimal, :integer], values: [a: 1, b: 2, c: 3]\n    field! :rating, :integer, when: category == target_category\n    field :score, :integer, derive: rating + category, gt: 1, lt: 100, when: score > rating\n  end\nend\n```\n\n```elixir\nTest.new!(%{category: 1, rating: 80}, target_category: 1)\n\n# %Test{category: 1, rating: 80, score: 81}\n```","ref":"Flint.Extensions.PreTransforms.html"},{"type":"function","title":"Flint.Extensions.PreTransforms.changeset/2","doc":"Applies transformations to each field according to the `:derive` options passed in the schema specification.\n\nThese transformations are applied after casting, but before validations when used within the default `Flint.Changeset.changeset` implementation.\n\nAccepts optional bindings which are passed to evaluated code.","ref":"Flint.Extensions.PreTransforms.html#changeset/2"},{"type":"module","title":"Flint.Extensions.Typed","doc":"Adds supports for **most** of the features from the wonderful [`typed_ecto_schema`](https://github.com/bamorim/typed_ecto_schema) library.\n\nRather than using the `typed_embedded_schema` macro from that library, however, thr `Typed` extension incorporates the features into the standard\n`embedded_schema` macro from `Flint.Schema`, meaning even fewer lines of code changed to use typed embedded schemas!\n\nIncluded with that are the addtional [Schema-Level options](https://hexdocs.pm/typed_ecto_schema/TypedEctoSchema.html#module-schema-options)\nyou can pass to the `embedded_schema` macro.\n\nYou also have the ability to [override field typespecs ](https://hexdocs.pm/typed_ecto_schema/TypedEctoSchema.html#module-overriding-the-typespec-for-a-field) as well as providing extra [field-level options](https://hexdocs.pm/typed_ecto_schema/TypedEctoSchema.html#module-extra-options) from\n`typed_ecto_schema`.\n\n> ##","ref":"Flint.Extensions.Typed.html"},{"type":"module","title":"Required vs Enforced vs Null {: .info} - Flint.Extensions.Typed","doc":"> Note that the typespecs allow you to specify `:enforce` and `:null` options, which are different from the requirement imposed by `field!`.\n> ---\n> `field!` marks the field as being required during the\n> changeset validation, which is equal to passing\n> the field name to the [`Eco.Changeset.validate_required/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_required/3).\n> An instance of the schema's struct **CAN** still be created if a field marked with `field!` is not present.\n> ---\n> `enforce: true` is equal to including that field in the [`@enforce_keys` module attribute](https://hexdocs.pm/elixir/structs.html#default-values-and-required-keys)\n> for the corresponding schema struct.\n>\n> An instance of the schema's struct **CANNOT** be created if a field marked with `enforce: true` is not present, as it\n> will `raise` an `ArgumentError`.\n> ---\n>`:null` indicates whether `nil` is a valid value for the field. This **only** impacts the generated typespecs.","ref":"Flint.Extensions.Typed.html#module-required-vs-enforced-vs-null-info"},{"type":"macro","title":"Flint.Extensions.Typed.embedded_schema/2","doc":"","ref":"Flint.Extensions.Typed.html#embedded_schema/2"},{"type":"module","title":"Flint.Extensions.When","doc":"The `When` extension adds the `:when` option to `Flint` schemas.\n\n`:when` lets you define an arbitrary boolean expression that will be evaluated and pass the validation if it\nevaluates to a truthy value. You may pass bindings to this condition and\nrefer to previously defined fields. `:when` also lets you refer to the current `field` in which\nthe `:when` condition is defined. Theoretically, you could write many of the other validations using `:when`, but you will\nreceive worse error messages with `:when` than with the dedicated validations.\n\n```elixir\ndefmodule Test do\n  use Flint.Schema\n\n  embedded_schema do\n    field! :category, Union, oneof: [Ecto.Enum, :decimal, :integer], values: [a: 1, b: 2, c: 3]\n    field! :rating, :integer, when: category == target_category\n    field! :score, :integer, gt: 1, lt: 100, when: score > rating\n  end\nend\n```\n\n```elixir\n> Test.new!(%{category: :a, rating: 80, score: 10}, target_category: :a)\n\n** (ArgumentError) %Test{category: :a, rating: 80, score: [\"Failed `:when` validation\"]}\n  (flint 0.0.1) lib/schema.ex:406: Flint.Schema.new!/3\n  (elixir 1.15.7) src/elixir.erl:396: :elixir.eval_external_handler/3\n  (stdlib 5.1.1) erl_eval.erl:750: :erl_eval.do_apply/7\n  (elixir 1.15.7) src/elixir.erl:375: :elixir.eval_forms/4\n  (elixir 1.15.7) lib/module/parallel_checker.ex:112: Module.ParallelChecker.verify/1\n  lib/livebook/runtime/evaluator.ex:622: anonymous fn/3 in Livebook.Runtime.Evaluator.eval/4\n  (elixir 1.15.7) lib/code.ex:574: Code.with_diagnostics/2\n```","ref":"Flint.Extensions.When.html"},{"type":"extras","title":"Flint Feature Walkthrough","doc":"# Flint Feature Walkthrough\n\n```elixir\nMix.install(\n  [\n    {:flint, \"~> 0.5\"},\n    {:poison, \"~> 6.0\"}\n  ],\n  consolidate_protocols: false\n)\n```","ref":"feature_guide.html"},{"type":"extras","title":"Introduction to Flint - Flint Feature Walkthrough","doc":"`Flint` is a library that aims to make `Ecto` `embedded_schema`'s more declarative, flexible, and expressive.\n\nOne of the core tenets of `Flint` is to be a drop-in replacement for `Ecto`, meaning that for all of your `Ecto` embedded schemas you can just switch from `use Ecto.Schema` to `use Flint.Schema`, and then you have to opt into all of the extra features of `Flint`.\n\n`Flint` aims to empower you to colocate as much information as possible into your `embedded_schema` definitions, but `Flint` core is very unopinionated about how you use that data. `Flint` provides several build in extensions that are opinionated, but fundamentally `Flint` just exposes a way for you to store and retrieve additional data in an `Ecto` `embedded_schema`.","ref":"feature_guide.html#introduction-to-flint"},{"type":"extras","title":"Core Features - Flint Feature Walkthrough","doc":"The core features of `Flint` are those not packaged as `Flint` extensions.\n\nAll of these features are available even when specifying `extensions: []` when using `Flint.Schema`\n\n- Added `field!`, `embeds_one!`, and `embeds_many!` macros that tag those fields as required and exposes them through the `__schema__(:required)` reflection function.\n- Added support for `do` block in `field` and `field!` to add `validation_condition -> error_message` pairs to the field.\n- Generated functions:\n  - `changeset/3`\n  - `new/2`\n  - `new!/2`\n- Custom implementation of `embedded_schema` that uses the above macros from `Flint.Schema` instead of `Ecto.Schema`.\n- The ability to define aliases for field options\n- Application-wide default options using configurations\n\nLet's explore each of these in more detail\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"feature_guide.html#core-features"},{"type":"extras","title":"Using Generated Functions - Flint Feature Walkthrough","doc":"### `changeset`, `new`, and `new!`\n\n<!-- livebook:{\"break_markdown\":true} -->\n\n`Flint` provides generated and overridable default implementations of `changeset`, `new`, and `new!` functions.\n\nThe `new` and `new!` functions use the `changeset` function.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nThe generated `changeset` function automatically accounts for the required fields, and now you can use `changeset` as you would any other changeset.\n\n**Many of the built-in extensions (which we will discuss in-depth later) build upon the core `changeset` function to build up a more comprehensive pipeline.**\n\nYou can also use the generated `new` and `new!` functions. `new` will create a new struct from the passed params and will apply the changes regardless of validation, as opposed to `new!`, which will `raise` on validation errors, but otherwise will apply any valid changes.\n\nLet's take a look at a more practical example. In this example, we're:\n\n- Using both normal and `!` variants of field declarations\n- Using the shorthand notation where we pass the schema as an option to the `use Flint` call\n- Using both external and inline `embeds` fields.\n- Using an `Ecto.Enum` field type to map values between `embedded` and `dumped` representations.\n\n```elixir\ndefmodule Book do\n  use Flint.Schema,\n    schema: [\n      field!(:title, :string),\n      embeds_one! :author, Author_d do\n        field!(:first_name)\n        field!(:last_name)\n        field(:bio, :string)\n      end,\n      embeds_many(:coauthors, Author_c),\n      field(:genre, Ecto.Enum, values: [biography: 0, science_fiction: 1, fantasy: 2, mystery: 3])\n    ]\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Book, <<70, 79, 82, 49, 0, 0, 32, ...>>, :ok}\n```\n\nNow when we call `Book.new` it will create a new `Book` struct regardless of validation errors.\n\n**Note that by `embeds_many(!)` fields will default to an empty list (`[]`) at all times, whereas `embeds_one!` defaults to `nil` as it marks the field as `:required`, whereas `embeds_one` defaults to the empty struct (of its embedding). You can control this behavior for `embeds_one` using the `defaults_to_struct` boolean option.**\n\n```elixir\nBook.new()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Book{title: nil, author: nil, coauthors: [], genre: nil}\n```\n\n```elixir\nBook.__schema__(:required)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:title, :author]\n```\n\n```elixir\nBook.new!(%{title: \"The old man and the sea\"})\n```\n\nThe generated `changeset` functions will also enforce `:required` validations for embedded fields, so if any required field of the `:author` embedded field is not present, then `Book` will fail validation.\n\n```elixir\nBook.new!(%{\n  title: \"Harry Potter\",\n  author: %{first_name: \"J.K.\"},\n  genre: :fantasy\n})\n```\n\n```elixir\nbook =\n  Book.new!(%{\n    title: \"Harry Potter\",\n    author: %{first_name: \"J.K.\", last_name: \"Rowling\"},\n    genre: :fantasy\n  })\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Book{\n  title: \"Harry Potter\",\n  author: %Book.Author_d{first_name: \"J.K.\", last_name: \"Rowling\", bio: nil},\n  coauthors: [],\n  genre: :fantasy\n}\n```\n\n```elixir\ndefmodule Book_b do\n  use Flint.Schema,\n    schema: [\n      field!(:title, :string),\n      embeds_one :author, Author_c do\n        field!(:first_name)\n        field!(:last_name)\n        field(:bio, :string)\n      end,\n      embeds_many(:coauthors, Author_d),\n      field(:genre, Ecto.Enum, values: [biography: 0, science_fiction: 1, fantasy: 2, mystery: 3])\n    ]\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Book_b, <<70, 79, 82, 49, 0, 0, 33, ...>>, :ok}\n```\n\nNote the quirk that if the embedded field is not marked as required, but one of its subfields is, then if the embedded struct is partially initialized it will fail if the required subfield is missing\n\n```elixir\nBook_b.new!(%{\n  title: \"Harry Potter\",\n  author: %{first_name: \"J.K.\"},\n  genre: :fantasy\n})\n```\n\nWhereas if it is not passed altogether then it will pass validation.\n\n```elixir\nBook_b.new!(%{\n  title: \"Harry Potter\",\n  genre: :fantasy\n})\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Book_b{\n  title: \"Harry Potter\",\n  author: %Book_b.Author_c{first_name: nil, last_name: nil, bio: nil},\n  coauthors: [],\n  genre: :fantasy\n}\n```","ref":"feature_guide.html#using-generated-functions"},{"type":"extras","title":"Required Fields - Flint Feature Walkthrough","doc":"```elixir\ndefmodule Person do\n  use Flint.Schema, extensions: []\n\n  @primary_key false\n  embedded_schema do\n    field! :first_name, :string\n    field :last_name, :string\n\n    embeds_many! :parents, Parent, primary_key: false do\n      field! :relationship, Ecto.Enum, values: [:mother, :father]\n      field! :first_name, :string\n      field :last_name, :string\n    end\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Person, <<70, 79, 82, 49, 0, 0, 21, ...>>, :ok}\n```\n\nNow we can use the generated function to create a new `Person` struct\n\n```elixir\nmark_twain = %{\n  first_name: \"Mark\",\n  last_name: \"Twain\",\n  parents: [\n    %{\n      first_name: \"John\",\n      last_name: \"Clemens\",\n      relationship: :father\n    },\n    %{\n      first_name: \"Jane\",\n      last_name: \"Clemens\",\n      relationship: :mother\n    }\n  ]\n}\n\nPerson.changeset(%Person{}, mark_twain)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Ecto.Changeset ,\n        valid?: true,\n        ...\n      >,\n      #Ecto.Changeset ,\n        valid?: true,\n        ...\n      >\n    ],\n    first_name: \"Mark\",\n    last_name: \"Twain\"\n  },\n  errors: [],\n  data: #Person<>,\n  valid?: true,\n  ...\n>\n```\n\nThis generates an `Ecto.Changeset`. Let's remove one of the fields that had been marked as required.\n\n```elixir\nmark_twain_bad = Map.drop(mark_twain, [:first_name])\nPerson.changeset(%Person{}, mark_twain_bad)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Ecto.Changeset ,\n        valid?: true,\n        ...\n      >,\n      #Ecto.Changeset ,\n        valid?: true,\n        ...\n      >\n    ],\n    last_name: \"Twain\"\n  },\n  errors: [first_name: {\"can't be blank\", [validation: :required]}],\n  data: #Person<>,\n  valid?: false,\n  ...\n>\n```\n\nThis time, we get an error in our changeset, since we no longer provided a `first_name`.\n\nYou can see this difference in the `new` and `new!` generated functions as well, which will automatically apply the changes from the changeset to produce a struct for the provided schema.\n\n```elixir\nPerson.new(mark_twain)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Person{\n  first_name: \"Mark\",\n  last_name: \"Twain\",\n  parents: [\n    %Person.Parent{id: nil, relationship: :father, first_name: \"John\", last_name: \"Clemens\"},\n    %Person.Parent{id: nil, relationship: :mother, first_name: \"Jane\", last_name: \"Clemens\"}\n  ]\n}\n```\n\n```elixir\nPerson.new(mark_twain_bad)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Person{\n  first_name: nil,\n  last_name: \"Twain\",\n  parents: [\n    %Person.Parent{id: nil, relationship: :father, first_name: \"John\", last_name: \"Clemens\"},\n    %Person.Parent{id: nil, relationship: :mother, first_name: \"Jane\", last_name: \"Clemens\"}\n  ]\n}\n```\n\nYou can see that with `new` we apply the changes regardless of whether there are any errors present in the changeset.\n\nIf we want to `raise` if an error is present then we can use `new!` instead.\n\n```elixir\nPerson.new!(mark_twain)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Person{\n  first_name: \"Mark\",\n  last_name: \"Twain\",\n  parents: [\n    %Person.Parent{id: nil, relationship: :father, first_name: \"John\", last_name: \"Clemens\"},\n    %Person.Parent{id: nil, relationship: :mother, first_name: \"Jane\", last_name: \"Clemens\"}\n  ]\n}\n```\n\n```elixir\nPerson.new!(mark_twain_bad)\n```\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"feature_guide.html#required-fields"},{"type":"extras","title":"Block Validations - Flint Feature Walkthrough","doc":"Block validations can be specified using `do` blocks in `field` and `field!`. These are specified as lists of `error_condition -> error_message` pairs. If the error condition returns `true`, then the corresponding `error_message` will be inserted into the changeset when using the generated `changeset`, `new`, and `new!` functions.\n\nWithin these validations, you can pass custom bindings, meaning that you can define these validations with respect to variables only available at runtime.\n\nIn addition to any bindings you pass, the calues of the fields themselves will be available as a variable with the same name as the field.\n\nYou can also refer to local and imported / aliased function within these validations as well.\n\n```elixir\ndefmodule Person do\n  use Flint.Schema, extensions: []\n\n  def starts_with_capital?(\"\"), do: false\n\n  def starts_with_capital?(< >) do\n    first in ?A..?Z\n  end\n\n  @primary_key false\n  embedded_schema do\n    field! :first_name, :string do\n      !starts_with_capital?(first_name) -> \"Must be capitalized!\"\n      String.length(first_name) >= 10 -> \"Name too long!\"\n    end\n\n    field(:last_name, :string)\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Person, <<70, 79, 82, 49, 0, 0, 23, ...>>, :ok}\n```\n\n```elixir\nPerson.new!(%{first_name: \"mark\"})\n```\n\n```elixir\nPerson.new!(%{first_name: \"Mark\"})\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Person{first_name: \"Mark\", last_name: nil}\n```\n\nAll error conditions will be checked, so if multiple error conditions are met then you can be sure that they are reflected in the changeset.\n\n```elixir\nPerson.new!(%{first_name: \"markmarkmark\"})\n```\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"feature_guide.html#block-validations"},{"type":"extras","title":"Extensions - Flint Feature Walkthrough","doc":"`Flint` is designed to be highly extensible and flexible. The main way to extend `Flint`s functionality is through extensions.\n\nFlint currently offers four ways to extend behavior:\n\n1. Schema-level attributes\n2. Field-level additional options\n3. Default `embedded_schema` definitions\n4. Injected Code\n\n**Note that extensions are inherited by every child embedded schema defined with `embeds_one` / `embeds_many`**\n\n<!-- livebook:{\"break_markdown\":true} -->\n\n#","ref":"feature_guide.html#extensions"},{"type":"extras","title":"1. Schema-Level Attributes - Flint Feature Walkthrough","doc":"<!-- livebook:{\"break_markdown\":true} -->\n\nExtension let you define schema-level attributes which can then be reflected upon later. This is a pattern already used in `Ecto` such as with the `@primary_key` attribute, which can then be retrieved with `__schema__(:primary_key)`. `Flint` simply lets you extend this to any module attribute you want.\n\nHowever, you can still use this to modify attributes already used by `Ecto`. Let's take a look at the built-in `Embedded` extension, which sets attributes to defaults which make more sense when using an `embedded_schema` rather than a `schema`, as an example.\n\n```elixir\ndefmodule Embedded do\n  use Flint.Extension\n\n  attribute :schema_prefix\n  attribute :schema_context\n  attribute :primary_key, default: false\n  attribute :timestamp_opts, default: [type: :naive_datetime]\nend\n\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Embedded, <<70, 79, 82, 49, 0, 0, 47, ...>>,\n [\n   Flint.Extension.Dsl.Attributes.Attribute.Options,\n   nil,\n   %{opts: [], entities: [%Flint.Extension.Field{...}, ...]}\n ]}\n```\n\nWhen you use the `Embedded` extension your schema will have these attributes set and can reflect on them.\n\n```elixir\ndefmodule ExampleEmbedded do\n  use Flint.Schema, extensions: [Embedded]\n\n  embedded_schema do\n    field :name, :string\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, ExampleEmbedded, <<70, 79, 82, 49, 0, 0, 20, ...>>, :ok}\n```\n\n```elixir\nExampleEmbedded.__schema__(:fields)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:name]\n```\n\nNotice that there is no `:id` field defined, which would be set if the `@primary_key` field were not set to `false`.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\n#","ref":"feature_guide.html#1-schema-level-attributes"},{"type":"extras","title":"2. Field-Level Additional Options - Flint Feature Walkthrough","doc":"<!-- livebook:{\"break_markdown\":true} -->\n\nThis built-in `JSON` extension is inspired by how marshalling is done in [Go](https://pkg.go.dev/encoding/json#Marshal), and registers three additional options that can be used to annotate a field. It then uses that option information to define implementations for `Jason` and `Poison` `Encoder` protocols, depending on which you specify.\n\nYou can see from that this uses the `__schema__(:extra_options)` reflection function on the schema, which stores all of the options registered across all extensions.\n\n```elixir\ndefmodule JSON do\n  use Flint.Extension\n\n  option :name, required: false, validator: &is_binary/1\n  option :omitempty, required: false, default: false, validator: &is_boolean/1\n  option :ignore, required: false, default: false, validator: &is_boolean/1\n\n  @doc false\n  def encode_to_map(module, struct) do\n    struct\n    |> Ecto.embedded_dump(:json)\n    |> Enum.reduce(%{}, fn {key, val}, acc ->\n      field_opts = get_field_options(module, key)\n      json_key = field_opts[:name] || to_string(key)\n\n      cond do\n        field_opts[:ignore] ->\n          acc\n\n        field_opts[:omitempty] && is_nil(val) ->\n          acc\n\n        true ->\n          Map.put(acc, json_key, val)\n      end\n    end)\n  end\n\n  defp get_field_options(module, field) do\n    module.__schema__(:extra_options)\n    |> Keyword.get(field, [])\n    |> Enum.into(%{})\n  end\n\n  defmacro __using__(opts) do\n    json_module = Keyword.get(opts, :json_module, Jason)\n    protocol = Module.concat([json_module, Encoder])\n\n    quote do\n      if Code.ensure_loaded?(unquote(json_module)) do\n        defimpl unquote(protocol) do\n          def encode(value, opts) do\n            encoded_map = Flint.Extensions.JSON.encode_to_map(unquote(__CALLER__.module), value)\n            unquote(Module.concat([json_module, Encoder, Map])).encode(encoded_map, opts)\n          end\n        end\n      end\n    end\n  end\nend\n\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, JSON, <<70, 79, 82, 49, 0, 0, 60, ...>>, {:__using__, 1}}\n```\n\nNow let's define a schema with specialized serialization using the `JSON` extension\n\n```elixir\ndefmodule ExampleJSON do\n  use Flint.Schema, extensions: [JSON]\n\n  @primary_key false\n  embedded_schema do\n    field! :first_name, :string, name: \"First Name\"\n    field :last_name, :string, name: \"Last Name\", omitempty: true\n    field :nicknames, {:array, :string}, ignore: true\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, ExampleJSON, <<70, 79, 82, 49, 0, 0, 21, ...>>, :ok}\n```\n\n```elixir\nperson =\n  ExampleJSON.new!(%{\n    first_name: \"Charles\",\n    nicknames: [\"Charlie\"]\n  })\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%ExampleJSON{first_name: \"Charles\", last_name: nil, nicknames: [\"Charlie\"]}\n```\n\n```elixir\nJason.encode!(person)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n\"{\\\"First Name\\\":\\\"Charles\\\"}\"\n```\n\nHere, you can see the effects of each of the 3 options:\n\n1. The `:first_name` field is encoded as `\"First Name\"` as specified with the `:name` option\n2. The `:last_name` field is not encoded at all, since it was empty (`nil`)\n3. The `:nicknames` field is not encoded since it is ignored\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nYou can also pass options to extensions. In this case, we can specify a different JSON library to use if we don't want to use `Jason`, which is the default.\n\n```elixir\ndefmodule ExamplePoison do\n  use Flint.Schema, extensions: [{JSON, json_module: Poison}]\n\n  @primary_key false\n  embedded_schema do\n    field! :first_name, :string, name: \"First Name\"\n    field :last_name, :string, name: \"Last Name\", omitempty: true\n    field :nicknames, {:array, :string}, ignore: true\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, ExamplePoison, <<70, 79, 82, 49, 0, 0, 21, ...>>, :ok}\n```\n\n```elixir\nperson =\n  ExamplePoison.new!(%{\n    first_name: \"Charles\",\n    nicknames: [\"Charlie\"]\n  })\nPoison.encode!(person)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n\"{\\\"First Name\\\":\\\"Charles\\\"}\"\n```\n\n#","ref":"feature_guide.html#2-field-level-additional-options"},{"type":"extras","title":"3. Default `embedded_schema` Definitions - Flint Feature Walkthrough","doc":"<!-- livebook:{\"break_markdown\":true} -->\n\nExtensions also let you define default `embedded_schema` definitions which will be merged with any schema that uses it.\n\n```elixir\ndefmodule Inherited do\n  use Flint.Extension\n\n  attribute(:schema_prefix)\n  attribute(:schema_context)\n  attribute(:primary_key, default: false)\n  attribute(:timestamp_opts, default: [type: :naive_datetime])\n\n  embedded_schema do\n    field!(:timestamp, :utc_datetime_usec)\n    field!(:id)\n\n    embeds_one :child, Child do\n      field(:name, :string)\n      field(:age, :integer)\n    end\n  end\nend\n\ndefmodule Schema do\n  use Flint.Schema, extensions: [Inherited]\n\n  embedded_schema do\n    field(:type, :string)\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Schema, <<70, 79, 82, 49, 0, 0, 23, ...>>, :ok}\n```\n\n```elixir\nSchema.__schema__(:fields)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:type, :timestamp, :id, :child]\n```\n\n```elixir\nSchema.__schema__(:embeds)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:child]\n```\n\n```elixir\ndefmodule Person do\n  use Flint.Schema\n\n  embedded_schema do\n    field :first_name\n    field :last_name\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Person, <<70, 79, 82, 49, 0, 0, 29, ...>>, :ok}\n```\n\n```elixir\ndefmodule Event do\n  use Flint.Extension\n\n  embedded_schema do\n    field!(:timestamp, :utc_datetime_usec)\n    field!(:id)\n    embeds_one(:person, Person)\n\n    embeds_one :child, Child do\n      field(:name, :string)\n      field(:age, :integer)\n    end\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Event, <<70, 79, 82, 49, 0, 0, 47, ...>>, :ok}\n```\n\n```elixir\ndefmodule Webhook do\n  use Flint.Schema, extensions: [Event, Embedded]\n\n  embedded_schema do\n    field :route, :string\n    field :name, :string\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Webhook, <<70, 79, 82, 49, 0, 0, 25, ...>>, :ok}\n```\n\n```elixir\nWebhook.__schema__(:extensions)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[Event, Flint.Extensions.Embedded]\n```\n\n```elixir\nWebhook.__schema__(:fields)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:route, :name, :timestamp, :id, :person, :child]\n```\n\n```elixir\nWebhook.__schema__(:embeds)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:person, :child]\n```\n\n#","ref":"feature_guide.html#3-default-embedded_schema-definitions"},{"type":"extras","title":"4. Injected Code - Flint Feature Walkthrough","doc":"<!-- livebook:{\"break_markdown\":true} -->\n\nLastly, extensions can define their own `__using__/1` macro that will be called by the schema using the extension.\n\nThis is also why the order in which extensions are specified matters, since extensions will be `use`d by the calling schema module in the order that they are specified in the `extensions` option.\n\nHere's an example of the built-in `Accessible` extensions, which implements the `Access` behaviour for the schema.\n\n```elixir\ndefmodule Accessible do\n  use Flint.Extension\n\n  defmacro __using__(_opts) do\n    quote do\n      @behaviour Access\n\n      @impl true\n      defdelegate fetch(term, key), to: Map\n      @impl true\n      defdelegate get_and_update(term, key, fun), to: Map\n      @impl true\n      defdelegate pop(data, key), to: Map\n    end\n  end\nend\n\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Accessible, <<70, 79, 82, 49, 0, 0, 48, ...>>, {:__using__, 1}}\n```\n\n```elixir\ndefmodule AccessibleSchema do\n  use Flint.Schema, extensions: [Accessible]\n\n  embedded_schema do\n    field :name\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, AccessibleSchema, <<70, 79, 82, 49, 0, 0, 22, ...>>, :ok}\n```\n\n```elixir\nperson = AccessibleSchema.new!(%{name: \"Mickey\"})\nperson[:name]\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n\"Mickey\"\n```\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"feature_guide.html#4-injected-code"},{"type":"extras","title":"Custom Types - Flint Feature Walkthrough","doc":"`Ecto` allows you to define custom types by implementing the [`Ecto.Type`](https://hexdocs.pm/ecto/Ecto.Type.html#module-custom-types-and-primary-keys) or [`Ecto.ParameterizedType`](https://hexdocs.pm/ecto/Ecto.ParameterizedType.html) behaviours.\n\nTypes are how you define the way in which data is imported (`cast` / `load`) and exported (`dump`) when using your schema. So when accepting external data, types are what determines how that data is put into and taken out of your struct, which is an important behavior to control when working ingesting external data or outputting your data to an external API.\n\nThis is common when using `Ecto` as a means to validate JSON data across programming language barriers.\n\nYou might find, however, as you try to write your own types that they can be quite tedious and verbose to implement.\n\nThat's where `Flint.Type` comes in!\n\n`Flint.Type` is meant to make writing new `Ecto` types require much less boilerplate, because you can base your\ntype off of an existing type and only modify the callbacks that have different behavior.\n\nSimply `use Flint.Type` and pass the `:extends` option which says which type module to inherit callbacks\nfrom. This will delegate all required callbacks and any implemented optional callbacks and make them\noverridable.\n\nIt also lets you make a type from an `Ecto.ParameterizedType` with default parameter values.\nYou may supply any number of default parameters. This essentially provides a new\n`init/1` implementation for the type, supplying the default values, while not affecting any of the\nother `Ecto.ParameterizedType` callbacks. You may still override the newly set defaults at the local level.\n\nJust supply all options that you wish to be defaults as extra options when using `Flint.Type`.\n\nYou may override any of the inherited callbacks inherit from the extended module\nin the case that you wish to customize the module further.\n\n```elixir\ndefmodule Category do\n  use Flint.Type, extends: Ecto.Enum, values: [:folder, :file]\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Category, <<70, 79, 82, 49, 0, 0, 12, ...>>, :ok}\n```\n\nThis will apply default `values` to `Ecto.Enum` when you supply a `Category` type\nto an Ecto schema. You may still override the values if you supply the `:values`\noption for the field.\n\n```elixir\ndefmodule Downloads do\n  use Flint.Schema\n\n  embedded_schema do\n    field :type, Category\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Downloads, <<70, 79, 82, 49, 0, 0, 29, ...>>, :ok}\n```\n\n```elixir\nDownloads.new!(%{type: :folder})\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Downloads{type: :folder}\n```\n\n```elixir\nDownloads.new!(%{type: :another})\n```\n\nThis will create a new `NewUID` type that behaves exactly like an `Ecto.UUID` except it dumps\nits string length.\n\n```elixir\nimport Flint.Type\ndeftype NewUID, extends: Ecto.UUID, dump: &String.length/1\n\ndefmodule TestType do\n  use Flint.Schema\n\n  embedded_schema do\n    field(:id, NewUID)\n  end\nend\n\nEcto.UUID.generate() |> NewUID.dump()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n36\n```\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"feature_guide.html#custom-types"},{"type":"extras","title":"Default Extensions - Flint Feature Walkthrough","doc":"You can get a list of the default extensions with `Flint.default_extensions()`.\n\n```elixir\nFlint.default_extensions()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[Flint.Extensions.PreTransforms, Flint.Extensions.When, Flint.Extensions.EctoValidations,\n Flint.Extensions.PostTransforms, Flint.Extensions.Accessible, Flint.Extensions.Embedded,\n Flint.Extensions.JSON]\n```\n\nYou can optionally provide an `:except` option to filter which extensions to use.\n\n```elixir\nFlint.default_extensions(except: [When])\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[Flint.Extensions.PreTransforms, Flint.Extensions.EctoValidations, Flint.Extensions.PostTransforms,\n Flint.Extensions.Accessible, Flint.Extensions.Embedded, Flint.Extensions.JSON]\n```\n\nWhen explicitly passing which extensions to use, the default extensions are not automatically included, so you can use `Flint.default_extensions` to use them in addition to whatever extensions you explicitly use.\n\n```elixir\ndefmodule MySchema do\n  use Flint.Schema, extensions: Flint.default_extensions(except: [JSON, When]) ++ [JSON]\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, MySchema, <<70, 79, 82, 49, 0, 0, 19, ...>>,\n {:module, Jason.Encoder.MySchema, <<70, 79, 82, ...>>, {:encode, 2}}}\n```\n\n```elixir\nMySchema.__schema__(:extensions)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[Flint.Extensions.PreTransforms, Flint.Extensions.EctoValidations, Flint.Extensions.PostTransforms,\n Flint.Extensions.Accessible, Flint.Extensions.Embedded, Flint.Extensions.JSON]\n```","ref":"feature_guide.html#default-extensions"},{"type":"extras","title":"Built-In Extensions - Flint Feature Walkthrough","doc":"`Flint` provides a bevy of built-in extensions (those listed in `Flint.default_extensions`) to provide common conveniences. When building out your own custom `Flint` extensions, you can refer to the implementation details for any of these extensions for reference.\n\nLet's walk through the different extensions:\n\n<!-- livebook:{\"break_markdown\":true} -->\n\n#","ref":"feature_guide.html#built-in-extensions"},{"type":"extras","title":"Accessible - Flint Feature Walkthrough","doc":"An extension to automatically implement the `Access` behaviour for your struct,\ndeferring to the `Map` implementation.\n\n```elixir\ndefmodule AccessibleSchema do\n  use Flint.Schema, extensions: [Accessible]\n\n  embedded_schema do\n    field :name\n    embeds_one :embedded, AccessibleEmbed do\n      field :type\n      field :category\n    end\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, AccessibleSchema, <<70, 79, 82, 49, 0, 0, 24, ...>>, :ok}\n```\n\n```elixir\na =\n  AccessibleSchema.new!(%{\n    name: \"SampleName\",\n    embedded: %{type: \"SampleType\", category: \"SampleCategory\"}\n  })\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%AccessibleSchema{\n  id: nil,\n  name: \"SampleName\",\n  embedded: %AccessibleSchema.AccessibleEmbed{\n    id: nil,\n    type: \"SampleType\",\n    category: \"SampleCategory\"\n  }\n}\n```\n\n```elixir\na[:name]\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n\"SampleName\"\n```\n\n```elixir\na[:embedded][:category]\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n\"SampleCategory\"\n```\n\n#","ref":"feature_guide.html#accessible"},{"type":"extras","title":"EctoValidations - Flint Feature Walkthrough","doc":"Shorthand options for common validations found in `Ecto.Changeset`\n\nJust passthrough the option for the appropriate validation and this extension\nwill take care of calling the corresponding function from `Ecto.Changeset` on\nyour data.\n\n##","ref":"feature_guide.html#ectovalidations"},{"type":"extras","title":"Options - Flint Feature Walkthrough","doc":"- `:greater_than` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n- `:less_than` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n- `:less_than_or_equal_to` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n- `:greater_than_or_equal_to` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n- `:equal_to` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n- `:not_equal_to` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n- `:format` ([`Ecto.Changeset.validate_format/4`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_format/4))\n- `:subset_of` ([`Ecto.Changeset.validate_subset/4`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_subset/4))\n- `:in` ([`Ecto.Changeset.validate_inclusion/4`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_inclusion/4))\n- `:not_in` ([`Ecto.Changeset.validate_exclusion/4`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_exclusion/4))\n- `:is` ([`Ecto.Changeset.validate_length/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_length/3-options))\n- `:min` ([`Ecto.Changeset.validate_length/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_length/3-options))\n- `:max` ([`Ecto.Changeset.validate_length/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_length/3-options))\n- `:count` ([`Ecto.Changeset.validate_length/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_length/3-options))\n\n##","ref":"feature_guide.html#options"},{"type":"extras","title":"Aliases - Flint Feature Walkthrough","doc":"By default, the following aliases are also available for convenience:\n\n<!-- livebook:{\"force_markdown\":true} -->\n\n```elixir\nconfig Flint, aliases: [\n  lt: :less_than,\n  gt: :greater_than,\n  le: :less_than_or_equal_to,\n  ge: :greater_than_or_equal_to,\n  eq: :equal_to,\n  ne: :not_equal_to\n]\n```\n\n```elixir\ndefmodule EctoValidationsSchema do\n  use Flint.Schema, extensions: [EctoValidations]\n\n  embedded_schema do\n    field! :first_name, :string,  max: 10, min: 5\n    field! :last_name, :string, min: 5, max: 10\n    field :favorite_colors, {:array, :string}, subset_of: [\"red\", \"blue\", \"green\"]\n    field! :age, :integer, greater_than: 0, less_than: max_age\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, EctoValidationsSchema, <<70, 79, 82, 49, 0, 0, 24, ...>>, :ok}\n```\n\n```elixir\nEctoValidationsSchema.changeset(\n  %EctoValidationsSchema{},\n  %{first_name: \"Bob\", last_name: \"Smith\", favorite_colors: [\"red\", \"blue\", \"pink\"], age: 101},\n  max_age: 100\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Ecto.Changeset ,\n  valid?: false,\n  ...\n>\n```\n\n#","ref":"feature_guide.html#aliases"},{"type":"extras","title":"Embedded - Flint Feature Walkthrough","doc":"An extension to house common default configurations for embedded schemas. These configurations are specific for\nin-memory schemas.\n\n##","ref":"feature_guide.html#embedded"},{"type":"extras","title":"Attributes - Flint Feature Walkthrough","doc":"The following attributes and defaults are set by this extension:\n\n- `:schema_prefix`\n- `:schema_context`\n- `:primary_key` - defaults to `false`\n- `:timestamp_opts` - defaults to `[type: :naive_datetime]`\n\nA new schema reflection function is made for each attribute:\n\n<!-- livebook:{\"force_markdown\":true} -->\n\n```elixir\n__schema__(:schema_context)\n...\n```\n\n```elixir\ndefmodule WithoutEmbedded do\n  use Flint.Schema, extensions: []\n\n  embedded_schema do\n    field :name\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, WithoutEmbedded, <<70, 79, 82, 49, 0, 0, 20, ...>>, :ok}\n```\n\n```elixir\nWithoutEmbedded.__schema__(:fields)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:id, :name]\n```\n\n```elixir\nWithoutEmbedded.__schema__(:primary_key)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:id]\n```\n\n```elixir\ndefmodule WithEmbedded do\n  use Flint.Schema, extensions: [Embedded]\n\n  embedded_schema do\n    field :name\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, WithEmbedded, <<70, 79, 82, 49, 0, 0, 20, ...>>, :ok}\n```\n\n```elixir\nWithEmbedded.__schema__(:fields)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:name]\n```\n\n```elixir\nWithEmbedded.__schema__(:primary_key)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[]\n```\n\n#","ref":"feature_guide.html#attributes"},{"type":"extras","title":"JSON - Flint Feature Walkthrough","doc":"Provides JSON encoding capabilities for Flint schemas with Go-like marshalling options.\n\nThis extension enhances Flint schemas with customizable JSON serialization options,\nallowing fine-grained control over how fields are represented in JSON output.\n\n##","ref":"feature_guide.html#json"},{"type":"extras","title":"Usage - Flint Feature Walkthrough","doc":"To use this extension, include it in your Flint schema:\n\n<!-- livebook:{\"force_markdown\":true} -->\n\n```elixir\n  defmodule MySchema do\n    use Flint.Schema,\n       extensions: [{JSON, json_module: :json}]  # Jason, or Poison\n       #extensions: [JSON] # (defaults to Jason if no args passed)\n\n    embedded_schema do\n      # Schema fields...\n    end\n  end\n```\n\n##","ref":"feature_guide.html#usage"},{"type":"extras","title":"JSON Encoding Options - Flint Feature Walkthrough","doc":"The following options can be specified for each field in your schema:\n\n- `:name` - Specifies a custom name for the field in the JSON output.\n- `:omitempty` - When set to `true`, omits the field from JSON output if its value is `nil`.\n- `:ignore` - When set to `true`, always excludes the field from JSON output.\n\n##","ref":"feature_guide.html#json-encoding-options"},{"type":"extras","title":"Defining Options - Flint Feature Walkthrough","doc":"Options are defined directly in your schema using the `field` macro:\n\n<!-- livebook:{\"force_markdown\":true} -->\n\n```elixir\n  embedded_schema do\n    field :id, :string, name: \"ID\"\n    field :title, :string, name: \"Title\", omitempty: true\n    field :internal_data, :map, ignore: true\n  end\n```\n\n```elixir\ndefmodule Book do\n  use Flint.Schema,\n    extensions: [Embedded, JSON]\n\n  embedded_schema do\n    field(:id, :string, name: \"ISBN\")\n    field(:title, :string)\n    field(:author, :string, omitempty: true)\n    field(:price, :decimal, name: \"SalePrice\")\n    field(:internal_notes, :string, ignore: true)\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Book, <<70, 79, 82, 49, 0, 0, 23, ...>>, :ok}\n```\n\n```elixir\nbook = %{\n  id: \"978-3-16-148410-0\",\n  title: \"Example Book\",\n  author: nil,\n  price: Decimal.new(\"29.99\"),\n  internal_notes: \"Not for customer eyes\"\n}\n\nbook |> Book.new!() |> Jason.encode!()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n\"{\\\"ISBN\\\":\\\"978-3-16-148410-0\\\",\\\"SalePrice\\\":\\\"29.99\\\",\\\"title\\\":\\\"Example Book\\\"}\"\n```\n\nYou can even specify an alternate JSON module from `Jason`, such as `Poison`. In reality, this works with any JSON library that uses a protocol with an `encoder` implementation to dispatch its JSON encoding. `Jason` and `Poison` are the only officially supported ones, both having been tested with the current implementation.\n\nYou can specify the JSON library like so:\n\n```elixir\ndefmodule PoisonBook do\n  use Flint.Schema,\n    extensions: [Embedded, {JSON, json_module: Poison}]\n\n  embedded_schema do\n    field(:id, :string, name: \"ISBN\")\n    field(:title, :string)\n    field(:author, :string, omitempty: true)\n    field(:price, :decimal, name: \"SalePrice\")\n    field(:internal_notes, :string, ignore: true)\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, PoisonBook, <<70, 79, 82, 49, 0, 0, 23, ...>>, :ok}\n```\n\n```elixir\nbook |> PoisonBook.new!() |> Poison.encode!()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n\"{\\\"title\\\":\\\"Example Book\\\",\\\"SalePrice\\\":29.99,\\\"ISBN\\\":\\\"978-3-16-148410-0\\\"}\"\n```\n\n#","ref":"feature_guide.html#defining-options"},{"type":"extras","title":"PreTransforms - Flint Feature Walkthrough","doc":"The `PreTransforms` provides a convenient `:derive` option to express how the field is computed.\n\n**By default, this occurs after casting and before validations.**\n\n`derived` fields let you define expressions with support for custom bindings to include any\n`field` declarations that occur before the current field.\n\n`:derive` will automatically put the result of the input expression into the field value.\nBy default, this occurs before any other validation, so you can still have access to `field`\nbindings and even the current computed field value (eg. within a `:when` validation from the\n`When` extension).\n\nYou can define a `derived` field with respect to the field itself, in which case it acts as\ntransformation. Typically in `Ecto`, incoming transformations of this support would happen\nat the `cast` step, which means the behavior is determined by the type in which you are casting into.\n`:derive` lets you apply a transformation after casting to change that behavior\nwithout changing the underlying allowed type.\n\nYou can also define a `derived` field with an expression that does not depend on the field,\nin which case it is suggested that you use the `field` macro instead of `field!` since any input\nin that case would be thrashed by the derived value. This means that a field can be completely\ndetermined as a product of other fields!\n\n```elixir\ndefmodule Test do\n  use Flint.Schema, extensions: [PreTransforms]\n\n  embedded_schema do\n    field!(:category, Union, oneof: [Ecto.Enum, :decimal, :integer], values: [a: 1, b: 2, c: 3])\n    field!(:rating, :integer)\n    field(:score, :integer, derive: rating + category)\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Test, <<70, 79, 82, 49, 0, 0, 24, ...>>, :ok}\n```\n\n```elixir\nTest.new!(%{category: 1, rating: 80})\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Test{id: nil, category: 1, rating: 80, score: 81}\n```\n\n#","ref":"feature_guide.html#pretransforms"},{"type":"extras","title":"PostTransforms - Flint Feature Walkthrough","doc":"The `PostTransforms` extension adds the `:map` option to `Flint` schemas.\n\nThis works similarly to the `PreTransforms` extension, but uses the `:map` option rather than\nthe `:derive` option used by `PreTransforms`, and by default, applies to the field after all validations.\n\nThe same caveats apply to the `:map` expression as all other expressions, with the exception that the\n`:map` function **only** accepts arity-1 anonymous functions or non-anonymous function expressions\n(eg. using variable replacement).\n\nIn the following example, `:derived` is used to normalize incoming strings to downcase to prepare for\nthe validation, then the output is mapped to the uppercase string using the `:map` option.\n\n```elixir\ndefmodule Character do\n  use Flint.Schema\n\n  embedded_schema do\n    field! :type, :string, derive: &String.downcase/1, map: String.upcase(type) do\n      type not in ~w[elf human] -> \"Expected elf or human, got: #{type}\"\n    end\n\n    field! :age, :integer do\n      age < 0 ->\n        \"Nobody can have a negative age\"\n\n      type == \"elf\" and age > max_elf_age ->\n        \"Attention! The elf has become a bug! Should be dead already!\"\n\n      type == \"human\" and age > max_human_age ->\n        \"Expected human to have up to #{max_human_age}, got: #{age}\"\n    end\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Character, <<70, 79, 82, 49, 0, 0, 32, ...>>, :ok}\n```\n\n```elixir\nmax_elf_age = 400\nmax_human_age = 120\nCharacter.new!(%{type: \"Elf\", age: 10}, binding())\n```\n\n#","ref":"feature_guide.html#posttransforms"},{"type":"extras","title":"When - Flint Feature Walkthrough","doc":"The `When` extension adds the `:when` option to `Flint` schemas.\n\n`:when` lets you define an arbitrary boolean expression that will be evaluated and pass the validation if it\nevaluates to a truthy value. You may pass bindings to this condition and\nrefer to previously defined fields. `:when` also lets you refer to the current `field` in which\nthe `:when` condition is defined. Theoretically, you could write many of the other validations using `:when`, but you will\nreceive worse error messages with `:when` than with the dedicated validations.\n\n```elixir\ndefmodule WhenTest do\n  use Flint.Schema\n\n  embedded_schema do\n    field!(:category, Union, oneof: [Ecto.Enum, :decimal, :integer], values: [a: 1, b: 2, c: 3])\n    field!(:rating, :integer, when: category == target_category)\n    field!(:score, :integer, gt: 1, lt: 100, when: score > rating)\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, WhenTest, <<70, 79, 82, 49, 0, 0, 31, ...>>, :ok}\n```\n\n```elixir\nWhenTest.new!(%{category: :a, rating: 80, score: 10}, target_category: :a)\n```\n\n<!-- livebook:{\"offset\":40528,\"stamp\":{\"token\":\"XCP._AM3hIZN9cnzdJ5ly9r32bXl5ns37vGCZ0Pi4oc8eytwBgs_rzFpWQNmJVUQD7x2w_M0OFFyZ5qdFpENg8Mqc5PpFLwYAZvp5Y0F44ZZh59SAAP2cywo\",\"version\":2}} -->","ref":"feature_guide.html#when"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,49,46,50]}}